#!/usr/bin/python
# -*- coding: utf-8 -*-
import sys
import socket


###############################################################################
########## Remote Bus Control Protocol (RBCP: UDP/IP) #########################
###############################################################################
class UDP(socket.socket):

	RBCP_VER_TYPE = 0xFF

	RBCP_CMDFLAG_READ_REQ  = 0xC0
	RBCP_CMDFLAG_READ_ACK  = 0xC8
	RBCP_CMDFLAG_WRITE_REQ = 0x80
	RBCP_CMDFLAG_WRITE_ACK = 0x88

	RBCP_ID = 7		# any
	RBCP_DATALEN = 2

	RBCP_PACKET_LEN_ADDR_ONLY = 8
	RBCP_PACKET_LEN_WITH_DATA = 10

	RBCP_SRCPORT_1ST = 60001
	RBCP_SRCPORT_END = 61000
	ISRUN_LITTLEENDIAN = True if sys.byteorder == 'little' else False

	##################################################################
	def __init__(self, defdest_ipaddr = None, defdest_port = None, src_ipaddr = '0.0.0.0', src_port = 0):
		super(UDP, self).__init__(socket.AF_INET, socket.SOCK_DGRAM)

		self.__readID = 0
		self.__writeID = 0
		self.__reqmsg = bytearray()
		self.__ackmsg = bytearray()
		self._defaultdestport = 4660
		self._defaultdestipaddr = '0.0.0.0'
		self.settimeout(0.2)	# set default timeout seconds

		if defdest_ipaddr is not None: self._defaultdestipaddr = defdest_ipaddr
		if defdest_port is not None: self._defaultdestport = defdest_port

		########## bind ##########
		if src_port > 0:
			port_zone = (src_port, src_port + 1)
		else:
			port_zone = (self.RBCP_SRCPORT_1ST, self.RBCP_SRCPORT_END + 1)

		lastEx = None
		binded = False

		for i in range(port_zone[0], port_zone[1]):
			try:
				self.bind((src_ipaddr, i))
				binded = True
				break
			except Exception as ex:
				lastEx = ex

		if binded == False: 
			raise lastEx

	##################################################################
	def getDestination(self, ipaddr = None, port = None):
		dest_ipaddr = ipaddr if ipaddr is not None else self._defaultdestipaddr
		dest_port = port if port is not None else self._defaultdestport
		return dest_ipaddr, dest_port

	##################################################################
	def __requestWrite(self, address, value, ipaddr = None, dport = None):
		dest_address = self.getDestination(ipaddr, dport)

		# send request
		self.__reqmsg = self.generateRBCPpacket(self.RBCP_CMDFLAG_WRITE_REQ, address, value)
		txbytes = self.sendto(self.__reqmsg, dest_address)

		if txbytes != len(self.__reqmsg):
			raise Exception('RBCP.__requestWrite(): TX Error...')

		# receive acknowledge
		read_bytes = self.RBCP_PACKET_LEN_WITH_DATA
		self.__ackmsg, srcaddr = self.recvfrom(read_bytes)
		if isinstance(self.__ackmsg, str) == True: self.__ackmsg = bytearray(self.__ackmsg)

		# check acknowledge
		if len(self.__ackmsg) != read_bytes:
			raise Exception('RBCP.__requestWrite(): acknowledge length error(%d - %d)...' % (read_bytes, len(self.__ackmsg)))

		check_msg = self.generateRBCPpacket(self.RBCP_CMDFLAG_WRITE_ACK, address, value)

		for i in range(len(check_msg)):
			if check_msg[i] != self.__ackmsg[i]:
				raise Exception('RBCP.__requestWrite(): acknowledge mismatch error(%d)...' % (i))


	##################################################################
	def __writeNbytes(self, address, txbytes, value, ipaddr = None, dport = None):
		value0 = bytearray(txbytes)

		for i in range(txbytes):
			value0[i] = (value >> (i * 8)) & 0x00ff

		# to network byteorder
		if self.ISRUN_LITTLEENDIAN == True:
			value0.reverse()

		for i in range( round(txbytes/2) ):
			self.__requestWrite(address + i*2, value0[i*2:i*2+2], ipaddr, dport)


	##################################################################
	def write2byte(self, address, value, ipaddr = None, dport = None): return self.__writeNbytes(address, 2, value, ipaddr, dport)
	def write4byte(self, address, value, ipaddr = None, dport = None): return self.__writeNbytes(address, 4, value, ipaddr, dport)
	def write6byte(self, address, value, ipaddr = None, dport = None): return self.__writeNbytes(address, 6, value, ipaddr, dport)
	def write8byte(self, address, value, ipaddr = None, dport = None): return self.__writeNbytes(address, 8, value, ipaddr, dport)

	##################################################################
	def __requestRead(self, address, ipaddr = None, dport = None):
		dest_address = self.getDestination(ipaddr, dport)

		# send request
		self.__reqmsg = self.generateRBCPpacket(self.RBCP_CMDFLAG_READ_REQ, address)
		txbytes = self.sendto(self.__reqmsg, dest_address)

		if txbytes != len(self.__reqmsg):
			raise Exception('RBCP.__requestRead(): TX Error...')

		# receive acknowledge
		rxbytes = self.RBCP_PACKET_LEN_WITH_DATA
		self.__ackmsg, srcaddr = self.recvfrom(rxbytes)
		self.__ackmsg = bytearray(self.__ackmsg)

		# check acknowledge
		if rxbytes != len(self.__ackmsg):
			raise Exception('RBCP.__requestRead(): acknowledge length error(%d - %d)...' % (rxbytes, len(self.__ackmsg)))

		check_msg = self.generateRBCPpacket(self.RBCP_CMDFLAG_READ_ACK, address, self.__ackmsg[8:rxbytes])

		for i in range(len(check_msg)):
			if check_msg[i] != self.__ackmsg[i]:
				raise Exception('RBCP.__requestRead(): acknowledge mismatch error(%d)...' % (i))

		return self.__ackmsg[8:rxbytes]

	##################################################################
	def __readNbytes(self, address, rxbytes, ipaddr = None, dport = None):
		value = 0
		read_cnt = round(rxbytes/2)
		for i in range(read_cnt):
			rxdata = self.__requestRead(address + i*2, ipaddr, dport)
			if self.ISRUN_LITTLEENDIAN == True: rxdata.reverse()

			wvalue = (( (rxdata[1] & 0x00FF)<<8 | (rxdata[0] & 0x00FF) ) & 0xFFFF)
			value += ( wvalue << (read_cnt-1-i)*16 )

		return value

	##################################################################
	def read2byte(self, address, ipaddr = None, dport = None): return self.__readNbytes(address, 2, ipaddr, dport)
	def read4byte(self, address, ipaddr = None, dport = None): return self.__readNbytes(address, 4, ipaddr, dport)
	def read6byte(self, address, ipaddr = None, dport = None): return self.__readNbytes(address, 6, ipaddr, dport)
	def read8byte(self, address, ipaddr = None, dport = None): return self.__readNbytes(address, 8, ipaddr, dport)

	##################################################################
	def generateRBCPpacket(self, cmd_flg, address, data = None):
		message = bytearray(8)
		message[0] = self.RBCP_VER_TYPE
		message[1] = cmd_flg
		message[2] = self.RBCP_ID
		message[3] = self.RBCP_DATALEN
		message[4] = (address >> 24) & 0x00ff
		message[5] = (address >> 16) & 0x00ff
		message[6] = (address >>  8) & 0x00ff
		message[7] = (address >>  0) & 0x00ff

		if data is not None:
			for i in range(len(data)):
				message.append(data[i])

		return message


###############################################################################
########## High-speed data transfer class (TCP/IP) ############################
###############################################################################
class TCP(socket.socket):

	##################################################################
	def __init__(self, timeout, ipaddr, port = 24, src_ipaddr = '0.0.0.0', src_port = 0):
		super(TCP, self).__init__(socket.AF_INET, socket.SOCK_STREAM)
		self.settimeout(timeout)
		self.bind((src_ipaddr, src_port))
		self.connect((ipaddr, port))

	##################################################################
	def read(self, rxbytes):

		sumbytes = 0
		rxdata = bytearray()

		if sys.version_info.major <= 2:		# python 2.#
			while rxbytes > sumbytes:
				try:
					read_bytes = rxbytes - sumbytes
					rxdata0 = self.recv(read_bytes)
				except Exception as ex:
					if ex.args[0] == 'timed out': rxdata0 = None
					else: raise ex

				if rxdata0 is not None:
					rxdata0 = bytearray(rxdata0)
					rxlength = len(rxdata0)

					if rxlength > 0:
						rxdata += rxdata0
						sumbytes += rxlength
					else: break
				else: break
		else:	# python 3.#~
			while rxbytes > sumbytes:
				try:
					read_bytes = rxbytes - sumbytes
					rxdata0 = self.recv(read_bytes)
				except Exception as ex:
					if ex.args[0] == 'timed out': rxdata0 = None
					else: raise ex

				if rxdata0 is not None:
					rxlength = len(rxdata0)

					if rxlength > 0:
						rxdata += rxdata0
						sumbytes += rxlength
					else: break
				else: break

		return rxdata, sumbytes

